@@ -18,11 +18,13 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 
+import difflib
 import logging
 import os
 import re
+import sys
 from filecmp import cmp as cmpfile
-from os.path import exists, isdir, islink, join, split, splitext
+from os.path import exists, isdir, islink, join, realpath, split, splitext
 from shutil import rmtree
 from stat import ST_MODE, S_IXUSR, S_IXGRP, S_IXOTH
 from dhpython import MULTIARCH_DIR_TPL
@@ -46,9 +48,10 @@ def fix_locations(package, interpreter, versions, options):
                 # TODO: what about relative symlinks?
                 log.debug('moving files from %s to %s', srcdir, dstdir)
                 share_files(srcdir, dstdir, interpreter, options)
-                parent_dir = '/'.join(srcdir.split('/')[:-1])
-                if exists(parent_dir) and not os.listdir(parent_dir):
-                    os.rmdir(parent_dir)
+                try:
+                    os.removedirs(srcdir)
+                except OSError:
+                    pass
 
         # do the same with debug locations
         dstdir = interpreter.sitedir(package, gdb=True)
@@ -56,45 +59,55 @@ def fix_locations(package, interpreter, versions, options):
             if isdir(srcdir):
                 log.debug('moving files from %s to %s', srcdir, dstdir)
                 share_files(srcdir, dstdir, interpreter, options)
-                parent_dir = '/'.join(srcdir.split('/')[:-1])
-                if exists(parent_dir) and not os.listdir(parent_dir):
-                    os.rmdir(parent_dir)
+                try:
+                    os.removedirs(srcdir)
+                except OSError:
+                    pass
 
 
 def share_files(srcdir, dstdir, interpreter, options):
     """Try to move as many files from srcdir to dstdir as possible."""
     for i in os.listdir(srcdir):
         fpath1 = join(srcdir, i)
+        if not exists(fpath1):  # removed in rename_ext
+            continue
         if not options.no_ext_rename and splitext(i)[-1] == '.so':
             # try to rename extension here as well (in :meth:`scan` info about
             # Python version is gone)
-            version = interpreter.parse_public_version(srcdir)
-            if version:
-                # note that if ver is empty, default Python version will be used
-                fpath1_orig = fpath1
-                new_name = interpreter.check_extname(i, version)
-                if new_name:
-                    fpath1 = join(srcdir, new_name)
-                if exists(fpath1):
-                    log.warn('destination file exist, '
-                             'cannot rename %s to %s', fpath1_orig, fpath1)
-                else:
-                    log.info('renaming %s to %s', fpath1_orig, fpath1)
-                    os.renames(fpath1_orig, fpath1)
-                    i = new_name
+            version = interpreter.parse_public_dir(srcdir)
+            if version and version is not True:
+                fpath1 = Scan.rename_ext(fpath1, interpreter, version)
+                i = split(fpath1)[-1]
         fpath2 = join(dstdir, i)
         if not isdir(fpath1) and not exists(fpath2):
             # do not rename directories here - all .so files have to be renamed first
             os.renames(fpath1, fpath2)
             continue
-        if isdir(fpath1):
+        if islink(fpath1):
+            # move symlinks without changing them if they point to the same place
+            if not exists(fpath2):
+                os.renames(fpath1, fpath2)
+            elif realpath(fpath1) == realpath(fpath2):
+                os.remove(fpath1)
+        elif isdir(fpath1):
             share_files(fpath1, fpath2, interpreter, options)
         elif cmpfile(fpath1, fpath2, shallow=False):
             os.remove(fpath1)
-        # XXX: check symlinks
+        else:
+            # The files differed so we cannot collapse them.
+            log.warn('Paths differ: %s and %s', fpath1, fpath2)
+            if options.verbose and not i.endswith('.so'):
+                with open(fpath1) as fp1:
+                    fromlines = fp1.readlines()
+                with open(fpath2) as fp2:
+                    tolines = fp2.readlines()
+                diff = difflib.unified_diff(fromlines, tolines, fpath1, fpath2)
+                sys.stderr.writelines(diff)
 
-    if exists(srcdir) and not os.listdir(srcdir):
-        os.rmdir(srcdir)
+    try:
+        os.removedirs(srcdir)
+    except OSError:
+        pass
 
 
 class Scan:
@@ -117,6 +130,7 @@ class Scan:
 
         self.options = options
         self.result = {'requires.txt': set(),
+                       'egg-info': set(),
                        'nsp.txt': set(),
                        'shebangs': set(),
                        'public_vers': set(),
@@ -130,11 +144,26 @@ class Scan:
                 del dirs[:]
                 continue
 
-            self.current_private_dir = None
-            self.current_pub_version = version = interpreter.parse_public_version(root)
-            if self.current_pub_version:  # i.e. a public site-packages directory
+            self.current_private_dir = self.current_pub_version = None
+            version = interpreter.parse_public_dir(root)
+            if version:
+                self.current_dir_is_public = True
+                if version is True:
+                    version = None
+                else:
+                    self.current_pub_version = version
+            else:
+                self.current_dir_is_public = False
+
+            if self.current_dir_is_public:
                 if root.endswith('-packages'):
-                    self.result['public_vers'].add(version)
+                    if version is not None:
+                        self.result['public_vers'].add(version)
+                    for name in dirs:
+                        if name in ('test', 'tests') or name.startswith('.'):
+                            log.debug('removing dist-packages/%s', name)
+                            rmtree(join(root, name))
+                            dirs.remove(name)
             else:
                 self.current_private_dir = self.check_private_dir(root)
                 if not self.current_private_dir:
@@ -183,7 +212,7 @@ class Scan:
                 fext = splitext(fn)[-1][1:]
                 if fext == 'so':
                     if not self.options.no_ext_rename:
-                        fpath = self.rename_ext(fpath)
+                        fpath = self.rename_ext(fpath, interpreter, version)
                     ver = self.handle_ext(fpath)
                     ver = ver or version
                     if ver:
@@ -208,6 +237,12 @@ class Scan:
                 if fext == 'py' and self.handle_public_module(fpath) is not False:
                     self.current_result['compile'] = True
 
+            if not dirs:
+                try:
+                    os.removedirs(root)
+                except OSError:
+                    pass
+
         log.debug("package %s details = %s", package, self.result)
 
     @property
@@ -222,7 +257,7 @@ class Scan:
     def is_unwanted_file(self, fpath):
         if self.__class__.UNWANTED_FILES.match(fpath):
             return True
-        if self.current_pub_version and self.is_dbg_package\
+        if self.current_dir_is_public and self.is_dbg_package\
                 and self.options.clean_dbg_pkg\
                 and splitext(fpath)[-1][1:] not in ('so', 'h'):
             return True
@@ -254,13 +289,17 @@ class Scan:
             if dpath.startswith(join('debian', self.package, i)):
                 return '/' + i
 
-    def rename_ext(self, fpath):
+    @staticmethod
+    def rename_ext(fpath, interpreter, current_pub_version=None):
         """Add multiarch triplet, etc. Return new name.
 
         This method is invoked for all .so files in public or private directories.
         """
+        # current_pub_version - version parsed from dist-packages (True if unversioned)
+        # i.e. if it's not None - it's a public dist-packages directory
+
         path, fname = fpath.rsplit('/', 1)
-        if self.current_pub_version and islink(fpath):
+        if current_pub_version is not None and islink(fpath):
             # replace symlinks with extensions in dist-packages directory
             dstfpath = fpath
             links = set()
@@ -278,7 +317,8 @@ class Scan:
         if MULTIARCH_DIR_TPL.match(fpath):
             # ignore /lib/i386-linux-gnu/, /usr/lib/x86_64-kfreebsd-gnu/, etc.
             return fpath
-        new_fn = self.interpreter.check_extname(fname, self.current_pub_version)
+
+        new_fn = interpreter.check_extname(fname, current_pub_version)
         if new_fn:
             # TODO: what about symlinks pointing to this file
             new_fpath = join(path, new_fn)
@@ -365,6 +405,7 @@ class Scan:
             else:
                 log.info('renaming %s to %s', name, clean_name)
                 os.rename(fpath, join(root, clean_name))
+        self.result['egg-info'].add(join(root, clean_name))
 
     def cleanup(self):
         if self.is_dbg_package and self.options.clean_dbg_pkg:
@@ -373,6 +414,6 @@ class Scan:
             for root, dirs, file_names in os.walk(proot, topdown=False):
                 if '-packages/' in root and not file_names:
                     try:
-                        os.rmdir(root)
+                        os.removedirs(root)
                     except Exception:
                         pass
